#############################################################################
##
#W  test.gi              GAP 4 package AtlasRep                 Thomas Breuer
##
#H  @(#)$Id: test.gi,v 1.78 2007/08/31 16:11:15 gap Exp $
##
#Y  Copyright (C)  2001,  Lehrstuhl D fuer Mathematik,  RWTH Aachen,  Germany
##
##  This file contains functions to test the data available in the
##  &ATLAS; of Group Representations.
##
Revision.( "atlasrep/gap/test_gi" ) :=
    "@(#)$Id: test.gi,v 1.78 2007/08/31 16:11:15 gap Exp $";


#############################################################################
##
#V  AtlasRepHardCases
##
InstallValue( AtlasRepHardCases, rec() );


#############################################################################
##
#F  AtlasOfGroupRepresentationsTestWords( [<tocid>[, <groupname>]][,]
#F                                        [<verbose>] )
##
AtlasRepHardCases.TestWords:= [
    [ "find", [ "B", "HN", "S417", "F24d2" ] ],
    [ "check", [ "B" ] ],
    [ "maxes", [ "Co1" ] ],
  ];

InstallGlobalFunction( AtlasOfGroupRepresentationsTestWords,
    function( arg )
    local result, tocid, verbose, types, toc, name, r, type, omit, entry;

    # Initialize the result.
    result:= true;

    if Length( arg ) = 0 then
      return AtlasOfGroupRepresentationsTestWords( "local", false );
    elif Length( arg ) = 1 and IsBool( arg[1] ) then
      return AtlasOfGroupRepresentationsTestWords( "local", arg[1] );
    elif Length( arg ) = 1 and IsString( arg[1] ) then
      return AtlasOfGroupRepresentationsTestWords( arg[1], false );
    elif Length( arg ) = 2 and IsString( arg[1] ) and IsString( arg[2] ) then
      return AtlasOfGroupRepresentationsTestWords( arg[1], arg[2], false );
    elif Length( arg ) = 2 and IsString( arg[1] ) and IsBool( arg[2] ) then
      for name in AtlasOfGroupRepresentationsInfo.groupnames do
        result:= AtlasOfGroupRepresentationsTestWords( arg[1],
                     name[3], arg[2] ) and result;
      od;
      return result;
    elif not ( Length( arg ) = 3 and IsString( arg[1] )
                                 and IsString( arg[2] )
                                 and IsBool( arg[3] ) ) then
      Error( "usage: AtlasOfGroupRepresentationsTestWords( [<tocid>[, ",
             "<groupname>]][,]\n[<verbose>] )" );
    fi;

    tocid:= arg[1];
    verbose:= arg[3];

    # Check only straight line programs.
    types:= AGRDataTypes( "prg" );

    toc:= AtlasTableOfContents( tocid );
    if toc = fail then
      # No test is reasonable.
      return true;
    fi;

    name:= First( AtlasOfGroupRepresentationsInfo.GAPnames,
                  x -> x[2] = arg[2] );
    if IsBound( toc.TableOfContents.( name[2] ) ) then
      r:= toc.TableOfContents.( name[2] );

      # Note that the ordering in the `and' statement must not be
      # changed, in order to execute all tests!
      for type in types do
        omit:= First( AtlasRepHardCases.TestWords,
                      pair -> pair[1] = type[1] );
        if IsBound( r.( type[1] ) ) then
          if IsList( omit ) and name[2] in omit[2] then
            if verbose then
              Print( "#I  omit TestWords for ", type[1], " and ", name[2],
                     "\n" );
            fi;
          else
            for entry in r.( type[1] ) do
              result:= type[2].TestWords( tocid, name[2],
                           entry[ Length( entry ) ], type, verbose )
                       and result;
            od;
          fi;
        fi;
      od;

    fi;

    # Return the result.
    return result;
end );


#############################################################################
##
#F  AtlasOfGroupRepresentationsTestFileHeaders( [<tocid>[,<groupname>]] )
##
InstallGlobalFunction( AtlasOfGroupRepresentationsTestFileHeaders,
    function( arg )
    local result, toc, record, type, entry, test, triple;

    # Initialize the result.
    result:= true;

    if Length( arg ) = 2 then

      toc:= AtlasTableOfContents( arg[1] );
      if toc = fail then
        # No test is reasonable.
        return true;
      fi;
      toc:= toc.TableOfContents;
      if IsBound( toc.( arg[2] ) ) then
        record:= toc.( arg[2] );
        for type in AGRDataTypes( "rep" ) do
          if IsBound( record.( type[1] ) ) then
            for entry in record.( type[1] ) do
              test:= type[2].TestFileHeaders( arg[1], arg[2], entry, type );
              if not IsBool( test ) then
                Print( "#E  ", test, " for ", entry[ Length( entry ) ],
                       "\n" );
                test:= false;
              fi;
              result:= test and result;
            od;
          fi;
        od;
      fi;

    elif Length( arg ) = 1 then

      for triple in AtlasOfGroupRepresentationsInfo.groupnames do
        result:= AtlasOfGroupRepresentationsTestFileHeaders( arg[1],
                     triple[3] ) and result;
      od;

    elif Length( arg ) = 0 then
      result:= AtlasOfGroupRepresentationsTestFileHeaders( "local" );
    fi;

    # Return the result.
    return result;
end );


#############################################################################
##
#F  AtlasOfGroupRepresentationsTestBinaryFormat()
##  
InstallGlobalFunction( AtlasOfGroupRepresentationsTestBinaryFormat,
    function()
    local tmpfile, formats, dir, result, filename, unpacked, mat;

    # Create one temporary file.
    tmpfile:= TmpName();

    # Get the filename format.
    formats:= Filtered(
        AtlasOfGroupRepresentationsInfo.TableOfContents.types.rep,
            x -> x[1] in [ "matff", "perm" ] );
    formats:= List( formats, format -> format[2].FilenameFormat );

    # Loop over the files in directory `dirname' that contain matrices.
    dir:= DirectoriesPackageLibrary( "atlasrep", "datagens" )[1];
    result:= true;
    for filename in DirectoryContents( Filename( dir, "" ) ) do
      unpacked:= filename;
      if Length( filename ) > 3 and filename{ [ Length( filename ) - 2
                                      .. Length( filename ) ] } = ".gz" then
        unpacked:= filename{ [ 1 .. Length( filename ) - 3 ] };
      fi;
      if ForAny( formats, format -> AGRParseFilenameFormat( unpacked,
                                        format ) <> fail ) then
        mat:= ScanMeatAxeFile( Filename( dir, unpacked ) );
        if mat = fail then
          Print( "#E  AtlasOfGroupRepresentationsTestBinaryFormat:\n",
                 "#E  corrupted file `", filename, "'\n" );
          result:= false;
        else
          CMtxBinaryFFMatOrPerm( mat, ConvertToMatrixRep( mat ), tmpfile );
          if mat <> FFMatOrPermCMtxBinary( tmpfile ) then
            Print( "#E  AtlasOfGroupRepresentationsTestBinaryFormat:\n",
                   "#E  differences for `", unpacked, "'\n" );
            result:= false;
          fi;
        fi;
      fi;
    od;

    # Remove the temporary file.
    RemoveFile( tmpfile );

    # Return the result.
    return result;
end );


#############################################################################
##
#F  AtlasOfGroupRepresentationsTestStandardization( [<gapname>] )
##
InstallGlobalFunction( AtlasOfGroupRepresentationsTestStandardization,
    function( arg )
    local result, name, gapname, groupname, toc, orders, nr, gens, std, ords;

    # Initialize the result.
    result:= true;

    if Length( arg ) = 0 then

      for name in AtlasOfGroupRepresentationsInfo.GAPnames do
        result:= AtlasOfGroupRepresentationsTestStandardization( name[1] )
                     and result;
      od;

    elif Length( arg ) = 1 and IsString( arg[1] ) then

      gapname:= arg[1];
      groupname:= First( AtlasOfGroupRepresentationsInfo.GAPnames,
                         pair -> pair[1] = gapname );
      if groupname = fail then
        Print( "#E  AtlasOfGroupRepresentationsTestStandardization:\n",
               "#E  no group with GAP name `", gapname, "'\n" );
        return false;
      fi;

      toc:= AtlasTableOfContents( "local" ).TableOfContents;

      orders:= [];

      # Loop over the relevant representations.
      nr:= 1;
      gens:= AtlasGenerators( gapname, nr );
      while gens <> fail do
        std:= gens.standardization;
        if ForAll( gens.generators, x -> Inverse( x ) <> fail ) then
          ords:= List( gens.generators, Order );
        else
          ords:= [ fail ];
        fi;
        if not ForAll( ords, IsInt ) then
          Print( "#E  representation `", gens.identifier,
                 "': non-finite order\n" );
          result:= false;
        elif IsBound( orders[ std ] ) then
          if orders[ std ] <> ords then
            Print( "#E  ", Ordinal( nr ), " representation of `", gapname,
                   "': incompatible orders ", ords, " and ", orders[ std ],
                   "\n" );
            result:= false;
          fi;
        else
          orders[ std ]:= ords;
        fi;
        nr:= nr + 1;
        gens:= AtlasGenerators( gapname, nr );
      od;

    fi;

    # Return the result.
    return result;
end );


#############################################################################
##
##  If the IO package is not installed then an error message is avoided
##  via the following assignment.
##
if not IsBound( IO_stat ) then
  IO_stat:= "dummy";
fi;


#############################################################################
##
#F  AtlasOfGroupRepresentationsTestTableOfContentsRemoteUpdates()
##
InstallGlobalFunction(
    AtlasOfGroupRepresentationsTestTableOfContentsRemoteUpdates, function()
    local version, inforec, home, server, path, dstfilename, result, lines,
          datadirs, line, pos, pos2, filename, localfile, servdate, stat;

    if LoadPackage( "io" ) <> true then
      Print( "#E  The package IO is not available.\n" );
      return fail;
    fi;

    # Download the file that lists the changes.
    version:= InstalledPackageVersion( "atlasrep" );
    inforec:= First( PackageInfo( "atlasrep" ), r -> r.Version = version );
    home:= inforec.PackageWWWHome;
    if home{ [ 1 .. 7 ] } = "http://" then
      home:= home{ [ 8 .. Length( home ) ] };
    fi;

    server:= home{ [ 1 .. Position( home, '/' ) - 1 ] };
    path:= home{ [ Position( home, '/' ) + 1 .. Length( home ) ] };
    dstfilename:= Filename( DirectoryTemporary(), "changes.htm" );

    result:= [];
    if AtlasOfGroupRepresentationsTransferFile( server,
               Concatenation( path, "/htm/data/changes.htm" ),
               dstfilename ) then
      lines:= SplitString( StringFile( dstfilename ), "\n" );
      lines:= Filtered( lines,
                  x ->     20 < Length( x ) and x{ [ 1 .. 4 ] } = "<tr>"
                       and x{ [ -3 .. 0 ] + Length( x ) } = " -->" );
      datadirs:= Concatenation(
                     DirectoriesPackageLibrary( "atlasrep", "datagens" ),
                     DirectoriesPackageLibrary( "atlasrep", "dataword" ) );
      for line in lines do
        pos:= PositionSublist( line, "</td><td>" );
        if pos <> fail then
          pos2:= PositionSublist( line, "</td><td>", pos );
          filename:= line{ [ pos+9 .. pos2-1 ] };
          localfile:= Filename( datadirs, filename );
          if localfile <> fail then
            if not IsExistingFile( localfile ) then
              localfile:= Concatenation( localfile, ".gz" );
            fi;
            if IsExistingFile( localfile ) then
              # There is something to compare.
              pos:= PositionSublist( line, "<!-- " );
              if pos <> fail then
                servdate:= Int( line{ [ pos+5 .. Length( line )-4 ] } );
                stat:= IO_stat( localfile );
                if stat <> fail then
                  if stat.mtime < servdate then
                    Add( result, localfile );
                  fi;
                fi;
              fi;
            fi;
          fi;
        fi;
      od;
    fi;
    return result;
end );

if IsString( IO_stat ) then
  Unbind( IO_stat );
fi;


#############################################################################
##
#F  AtlasOfGroupRepresentationsTestFiles( [<tocid>[, <groupname>]] )
##
InstallGlobalFunction( AtlasOfGroupRepresentationsTestFiles, function( arg )
    local result, toc, record, type, entry, triple;

    # Initialize the result.
    result:= true;

    if Length( arg ) = 2 then
      toc:= AtlasTableOfContents( arg[1] );
      if toc = fail then
        return false;
      fi;
      toc:= toc.TableOfContents;
      if IsBound( toc.( arg[2] ) ) then
        record:= toc.( arg[2] );
        for type in AGRDataTypes( "rep" ) do
          if IsBound( record.( type[1] ) ) then
            for entry in record.( type[1] ) do
              result:= type[2].TestFiles( arg[1], arg[2], entry, type )
                       and result;
            od;
          fi;
        od;
      fi;
    elif Length( arg ) = 1 then
      for triple in AtlasOfGroupRepresentationsInfo.groupnames do
        result:= AtlasOfGroupRepresentationsTestFiles( arg[1], triple[3] )
                 and result;
      od;
    elif IsEmpty( arg ) then
      result:= AtlasOfGroupRepresentationsTestFiles( "local" ) and result;
    fi;

    # Return the result.
    return result;
end );


#############################################################################
##
#F  AtlasOfGroupRepresentationsTestClassScripts()
#F  AtlasOfGroupRepresentationsTestClassScripts( <groupname> )
##
##  The following groups are (currently) too large for checking
##  centralizer orders in a concrete representation.
#T  Replace this by the check for a suff. small permutation representation?
##
AtlasRepHardCases.TestClassScripts:= [
    "B", "F24", "HN", "J4", "Ly", "M", "ON", "Th", "L37d2",
  ];

InstallGlobalFunction( AtlasOfGroupRepresentationsTestClassScripts,
    function( arg )
    local result, groupname, gapname, toc, record, std, name, prg, tbl,
          outputs, ident, classnames, map, gens, roots, grp, reps, orders1,
          orders2, cents1, cents2, triple, pos, pos2, cycscript;

    # Initialize the result.
    result:= true;

    if Length( arg ) = 1 and IsString( arg[1] ) then

      groupname:= arg[1];
      gapname:= First( AtlasOfGroupRepresentationsInfo.GAPnames,
                       pair -> pair[2] = groupname );
      if gapname = fail then
        Print( "#E  no group with name `", groupname, "'\n" );
        return false;
      fi;
      gapname:= gapname[1];
      toc:= AtlasTableOfContents( "local" );
      if toc = fail then
        return false;
      fi;
      toc:= toc.TableOfContents;
#T admit also private tables of contents!
      if IsBound( toc.( groupname ) ) then
        record:= toc.( groupname );
        for name in [ "cyclic", "classes", "cyc2ccl" ] do
          if IsBound( record.( name ) ) then
            for std in Set( List( record.( name ), x -> x[1] ) ) do

              prg:= AtlasProgram( gapname, std, name );
              if prg = fail then
                Print( "#E  inconsistent program `", name, "' for `",
                       gapname, "'\n" );
                result:= false;
              else

                # Fetch the character table of the group.
                # (No further tests are possible if it is not available.)
                tbl:= CharacterTable( gapname );
                if tbl <> fail then

                  ident:= prg.identifier[2];
                  classnames:= AtlasClassNames( tbl );
                  if IsBound( prg.outputs ) then
                    outputs:= prg.outputs;
                    map:= List( outputs, x -> Position( classnames, x ) );
                  else
                    Print( "#E  no component `outputs' in `", name,
                           "' for `", gapname, "'\n" );
                    result:= false;
                    outputs:= [ "-" ];
                    map:= [ fail ];
                  fi;
                  prg:= prg.program;

                  # (If `-' signs occur then we cannot test the names,
                  # but the number of outputs can be checked.)
                  roots:= ClassRoots( tbl );
                  roots:= Filtered( [ 1 .. Length( roots ) ],
                                    i -> IsEmpty( roots[i] ) );
                  roots:= Set( List( roots, x -> ClassOrbit( tbl, x ) ) );

                  if ForAll( outputs, x -> not '-' in x ) then

                    # Check the class names.
                    if fail in map then
                      Print( "#E  strange class names ",
                             Difference( outputs, classnames ),
                             " for `dataword/", ident, "'\n" );
                      result:= false;
                    fi;
                    if     name in [ "classes", "cyc2ccl" ]
                       and Set( classnames ) <> Set( outputs ) then
                      Print( "#E  class names ",
                             Difference( classnames, outputs ),
                             " not hit for `dataword/", ident, "'\n" );
                      result:= false;
                    fi;
                    if name = "cyclic" then
                      # Check whether all maximally cyclic subgroups
                      # are covered.
                      roots:= Filtered( roots,
                                 list -> IsEmpty( Intersection( outputs,
                                             classnames{ list } ) ) );
                      if not IsEmpty( roots ) then
                        Print( "#E  maximally cyclic subgroups ",
                               List( roots, x -> classnames{ x } ),
                               " not hit for `dataword/", ident, "'\n" );
                        result:= false;
                      fi;
                    fi;

                  elif name = "cyclic" and
                       Length( outputs ) <> Length( roots ) then
                    Print( "#E  no. of outputs and cyclic subgroups differ",
                           " for `dataword/", ident, "'\n" );
                  fi;

                  if not ( fail in map or
                       groupname in AtlasRepHardCases.TestClassScripts ) then

                    # Compute the representatives in a representation.
                    # (No further tests are possible if none is available.)
                    gens:= OneAtlasGeneratingSetInfo( gapname, std,
                                                      IsPermGroup, true );
                    if gens <> fail then

                      gens:= AtlasGenerators( gens.identifier );
                      if gens <> fail then
                        gens:= gens.generators;
                      fi;
                      if fail in gens then
                        gens:= fail;
                      fi;

                      if not name in [ "cyclic", "classes" ] then

                        # The input consists of the images of the standard
                        # generators under the `cyc' script.
                        pos:= Position( ident, '-' ) - 1;
                        pos2:= pos;
                        while ident[ pos2 ] <> 'W' do
                          pos2:= pos2 - 1;
                        od;
                        cycscript:= Concatenation( groupname, "G",
                                        String( std ), "-cycW",
                                        ident{ [ pos2+1 .. pos ] } );
                        cycscript:= AtlasProgram(
                            [ gapname, cycscript, std ] );
                        if cycscript = fail then
                          gens:= fail;
                          Print( "#E  no script `", cycscript,
                                 "' available\n" );
                          result:= false;
                        else
                          gens:= ResultOfStraightLineProgram(
                                     cycscript.program, gens );
                        fi;
                      fi;

                    fi;

                    if gens <> fail then

                      grp:= Group( gens );
                      reps:= ResultOfStraightLineProgram( prg, gens );

                      if Length( reps ) <> Length( outputs ) then

                        Print( "#E  inconsistent output numbers for ",
                               "`dataword/", ident, "'\n" );
                        result:= false;

                      else

                        # Check element orders and centralizer orders.
                        orders1:= OrdersClassRepresentatives( tbl ){ map };
                        orders2:= List( reps, Order );
                        if orders1 <> orders2 then
                          Print( "#E  element orders of ",
                              outputs{ Filtered( [ 1 .. Length( outputs ) ],
                                           i -> orders1[i] <> orders2[i] ) },
                              " differ for `dataword/", ident, "'\n" );
                          result:= false;
                        fi;
                        cents1:= SizesCentralizers( tbl ){ map };
                        cents2:= List( reps, x -> Size( Centralizer(grp,x) ) );
                        if    cents1 <> cents2 then
                          Print( "#E  centralizer orders of ",
                              outputs{ Filtered( [ 1 .. Length( outputs ) ],
                                           i -> cents1[i] <> cents2[i] ) },
                              " differ for `dataword/", ident, "'\n" );
                          result:= false;
                        fi;

                      fi;

                    fi;

                  fi;

                fi;
              fi;

            od;
          fi;
        od;
      fi;

    elif IsEmpty( arg ) then
      for triple in AtlasOfGroupRepresentationsInfo.groupnames do
        result:= AtlasOfGroupRepresentationsTestClassScripts( triple[3] )
                 and result;
      od;
    fi;

    # Return the result.
    return result;
end );


#############################################################################
##
#F  AtlasOfGroupRepresentationsTestCycToCcls( [<groupname>] )
##
InstallGlobalFunction( AtlasOfGroupRepresentationsTestCycToCcls,
    function( arg )
    local result, groupname, gapname, toc, tbl, record, datadirs, entry,
          tomatch, cyc2ccl, str, prg, triple;

    # Initialize the result.
    result:= true;

    if Length( arg ) = 1 and IsString( arg[1] ) then

      groupname:= arg[1];
      gapname:= First( AtlasOfGroupRepresentationsInfo.GAPnames,
                       pair -> pair[2] = groupname );
      if gapname = fail then
        Print( "#E  no group with name `", groupname, "'\n" );
        return false;
      fi;
      gapname:= gapname[1];
      toc:= AtlasTableOfContents( "local" );
      if toc = fail then
        return false;
      fi;
      toc:= toc.TableOfContents;

      # Fetch the character table of the group.
      # (No test is possible if it is not available.)
      tbl:= CharacterTable( gapname );
      if   tbl = fail then
        Print( "#I  no character table of `", gapname, "' is available\n" );
        return true;
      elif not IsBound( toc.( groupname ) ) then
        return true;
      fi;

      record:= toc.( groupname );
      if IsBound( record.cyclic ) then
        if IsBound( record.cyc2ccl ) then
          cyc2ccl:= List( record.cyc2ccl, x -> SplitString( x[2], "-" ) );
        else
          cyc2ccl:= [];
        fi;

        datadirs:= DirectoriesPackageLibrary( "atlasrep", "dataword" );

        for entry in record.cyclic do

          # Check the `cyc2ccl' scripts available.
          tomatch:= Filtered( entry[2], x -> x <> '-' );
          cyc2ccl:= Filtered( cyc2ccl, x -> x[1] = tomatch );
          if IsEmpty( cyc2ccl ) then

            # There is no `cyc2ccl' script but perhaps we can create it.
            str:= StringOfAtlasProgramCycToCcls(
                      Filename( datadirs, entry[2] ), tbl, "names" );
            if str <> fail then
              prg:= ScanStraightLineProgram( str, "string" );
              if prg = fail then
                Print( "#E  automatically created script for `", tomatch,
                       "-cclsW1' would be incorrect" );
              fi;
              prg:= prg.program;
#T check the composition?
              Print( "#I  add the following script, in the new file `",
                     tomatch, "-cclsW1':\n",
                     str, "\n" );
              result:= false;
            fi;

          fi;
        od;
      fi;

    elif IsEmpty( arg ) then
      for triple in AtlasOfGroupRepresentationsInfo.groupnames do
        result:= AtlasOfGroupRepresentationsTestCycToCcls( triple[3] )
                 and result;
      od;
    fi;

    # Return the result.
    return result;
end );


#############################################################################
##
#F  AtlasOfGroupRepresentationsTestGroupOrders( [true] )
##
InstallGlobalFunction( AtlasOfGroupRepresentationsTestGroupOrders,
    function( arg )
    local verbose, formats, SizesFromName, result, entry, size;

    verbose:= ( Length( arg ) <>  0 and arg[1] = true );

    formats:= [
      [ [ "L", IsDigitChar, "(", IsDigitChar, ")" ],
        l -> Size( PSL( l[2], l[4] ) ) ],
      [ [ "2.L", IsDigitChar, "(", IsDigitChar, ")" ],
        l -> 2 * Size( PSL( l[2], l[4] ) ) ],
      [ [ "S", IsDigitChar, "(", IsDigitChar, ")" ],
        l -> Size( PSp( l[2], l[4] ) ) ],
      [ [ "2.S", IsDigitChar, "(", IsDigitChar, ")" ],
        l -> 2 * Size( PSp( l[2], l[4] ) ) ],
      [ [ "U", IsDigitChar, "(", IsDigitChar, ")" ],
        l -> Size( PSU( l[2], l[4] ) ) ],
    ];

    SizesFromName:= function( name )
      local result, pair, parse, tbl, tom, data, pos, size1, size2;

      result:= [];

      # Deal with the case of integers.
      if ForAll( name, x -> IsDigitChar( x ) or x = '^' ) then
#T improve: admit also brackets and '+' (problem of *matching* brackets)
        # No other criterion matches with this format, so we return.
        return [ EvalString( name ) ];
      fi;

      for pair in formats do
        parse:= ParseBackwards( name, pair[1] );
        if parse <> fail then
          AddSet( result, pair[2]( parse ) );
        fi;
      od;

      tbl:= CharacterTable( name );
      if tbl <> fail then
        AddSet( result, Size( tbl ) );
      fi;
      tom:= TableOfMarks( name );
      if tom <> fail then
        AddSet( result, Size( UnderlyingGroup( tom ) ) );
      fi;

      # Try to use the database.
      data:= OneAtlasGeneratingSetInfo( name, NrMovedPoints, [ 1 .. 10^4 ] );
   #  if data = fail then
   #    data:= OneAtlasGeneratingSetInfo( name,
   #               Dimension, [ 1 .. 10 ] );
   #  fi;
      if data <> fail then
        data:= AtlasGenerators( data );
        if data <> fail then
          AddSet( result, Size( Group( data.generators ) ) );
        fi;
      fi;

      # Try to evaluate the name structure.
      pos:= Position( name, '.' );
#T improve: split also at ':'
      if pos <> fail then
        size1:= SizesFromName( name{ [ 1 .. pos-1 ] } );
        size2:= SizesFromName( name{ [ pos+1 .. Length( name ) ] } );
        if Length( size1 ) = 1 and Length( size2 ) = 1 then
          AddSet( result, size1[1] * size2[1] );
        elif Length( size1 ) > 1 or Length( size2 ) > 1 then
          Print( "#E  group orders: problem with `", name, "'\n" );
        fi;
      fi;

      return result;
    end;

    result:= true;

    for entry in AtlasOfGroupRepresentationsInfo.GAPnames do
      size:= SizesFromName( entry[1] );
      if Length( entry ) = 2 then
        if Length( size ) <> 1 then
          Print( "#I  AGRGNAN: group order for `", entry[1], "' unknown\n" );
        else
          Add( entry, size[1] );
          Print( "#I  AGRGNAN: set group order `", size[1], "' for `",
                 entry[1], "'\n" );
        fi;
      elif Length( size ) = 0 then
        if verbose then
          Print( "#I  AGRGNAN: cannot verify group order for `", entry[1],
                 "'\n" );
        fi;
      elif 1 < Length( size ) then
        Print( "#E  AGRGNAN: several group orders for `", entry[1],
               "':\n#E  ", size, "\n" );
        result:= false;
      elif size[1] <> entry[3] then
        Print( "#E  AGRGNAN: wrong group order for `", entry[1], "'\n" );
        result:= false;
      fi;
    od;

    return result;
end );


#############################################################################
##
#E

