############################################################################
##
#W  lpres.gi 			NQL				Rene Hartung
##
#H   @(#)$Id: lpres.gi,v 1.5 2007/04/23 12:56:45 gap Exp $
##
## Based on Alexander Hulpke's and Volkmar Felsch's construction for 
## finitely presented groups ("GAPDIR/lib/grpfp.gi").
##

############################################################################
##
#M  ElementOfLpGroup
##
InstallMethod( ElementOfLpGroup,
   "for a family of L-presented group elements, and an assoc. word",
   true,
   [ IsElementOfLpGroupFamily, IsAssocWordWithInverse ],
   function( fam, elm )
     return Objectify(fam!.defaultType, [Immutable(elm)]);
   end);

############################################################################
##
#M  PrintObj( <elm> ) 
## 
InstallMethod( PrintObj,
  "for an element of an L-presented group",
  true,
  [IsElementOfLpGroup and IsPackedElementDefaultRep], 0,
  function(obj)
    Print(obj![1]);
  end);

############################################################################
##
#M  ViewObj( <elm> ) 
## 
InstallMethod( ViewObj,
  "for an element of an L-presented group",
  true,
  [IsElementOfLpGroup and IsPackedElementDefaultRep], 0,
  function(obj)
    View(obj![1]);
  end);

############################################################################
##
#M  String( <elm> ) 
## 
InstallMethod( String,
  "for an element of an L-presented group",
  true,
  [IsElementOfLpGroup and IsPackedElementDefaultRep], 0,
  obj-> String(obj![1]));

############################################################################
##
#M  LaTeXObj( <elm> ) 
## 
InstallMethod( LaTeXObj,
  "for an element of an L-presented group",
  true,
  [IsElementOfLpGroup and IsPackedElementDefaultRep], 0,
  obj->LaTeXObj(obj![1]));

############################################################################
##
#M  UnderlyingElement( <elm> ) 
## 
InstallMethod( UnderlyingElement,
  "for an element of an L-presented group",
  true,
  [IsElementOfLpGroup and IsPackedElementDefaultRep], 0,
   obj -> obj![1]);

############################################################################
##
#M  ExtRepOfObj( <elm> )
## 
InstallMethod( ExtRepOfObj,
  "for an element of an L-presented group",
  true,
  [IsElementOfLpGroup and IsPackedElementDefaultRep], 0,
  obj -> ExtRepOfObj(obj![1]));

############################################################################
##
#M  Length
##
InstallOtherMethod( Length, 
  "for an element of an L-presented group", 
  true,
  [IsElementOfLpGroup and IsPackedElementDefaultRep], 0,
  x->Length(UnderlyingElement(x)));

############################################################################
##
#M  InverseOp ( <elm> ) 
## 
InstallMethod( InverseOp,
  "for an element of an L-presented group",
  true,
  [IsElementOfLpGroup], 0,
  obj->ElementOfLpGroup(FamilyObj(obj),Inverse(UnderlyingElement(obj))));

############################################################################
##
#M  One( <fam> )
##
InstallOtherMethod( One,
  "for a family of L-presented group elements",
  true,
  [IsElementOfLpGroupFamily],0,
  fam->ElementOfLpGroup(fam,One(fam!.freeGroup)));

############################################################################
##
#M  One( <elm> )
##
InstallMethod( One, 
  "for an L-presented group element", 
  true,
  [IsElementOfLpGroup], 0, 
  obj->One(FamilyObj(obj)));

#a^0 calls OneOp
InstallMethod( OneOp, 
  "for an L-presented group element",
  true,
  [IsElementOfLpGroup],0,
  obj -> One(FamilyObj(obj)));

#############################################################################
##
#M  \*( <elm1>, <elm2> )  . . . . . for two elements of an L-presented group 
##
InstallMethod( \*, 
   "for two L-presented group elements",
    IsIdenticalObj, 
    [ IsElementOfLpGroup, IsElementOfLpGroup ], 0,
    function( left, right )
    local fam,k;
    fam:= FamilyObj( left );

    return ElementOfLpGroup( fam,
      UnderlyingElement( left ) * UnderlyingElement( right ) );
end );

############################################################################
##
#M  MappedWord ( word, gens, imgs )
## 
##  Replaced each occurence of gen from gens in word by img in imgs.
##
InstallOtherMethod( MappedWord,
   "for LpGroup elements",
   true,
   [IsElementOfLpGroup,IsList,IsList], 0,
   function(x,gens1,gens2)
   if not IsElementOfLpGroupCollection(gens1) then
     return(fail);
   else
     return(MappedWord(UnderlyingElement(x),
                       List(gens1,UnderlyingElement),gens2));
   fi;
end);

############################################################################
##
#M  GeneratorsOfGroup( <F> )  . . . . . . . . . . . for an L-presented group
##
InstallMethod( GeneratorsOfGroup, 
   "for whole family of L-presented groups", 
   true,
   [ IsLpGroup and IsGroupOfFamily ], 0,
   function( F )
   local Fam;	
   Fam:= ElementsFamily( FamilyObj( F ) );
   return List( FreeGeneratorsOfLpGroup( F ), g -> ElementOfLpGroup( Fam, g ) );
   end );

############################################################################
##
#M  Display( <G> ) . . . . . . . . . . . . . . . . . . .  display an LpGroup
##
InstallMethod( Display,
    "for L-presented groups",
    true,
    [ IsLpGroup and IsGroupOfFamily ],
    0,
    function( G )
    local   gens,       # generators o the free group
            rels,       # relators of <G>
            endos,      # endomorphisms of <G>
            itrels,     # iterated relators of <G>
            n,          # number of relators, endomorphisms and it. relators
            i;          # loop variable

    # generators of the L-presentation
    gens := FreeGeneratorsOfLpGroup( G );
    Print( "generators = ", gens, "\n" );

    # fixed relators of the L-presentation
    rels := FixedRelatorsOfLpGroup( G );
    n := Length( rels );
    Print( "fixed relators = [" );
    if n > 0 then
        Print( "\n ", rels[1] );
        for i in [ 2 .. n ] do
          Print( ",\n ", rels[i] );
        od;
    fi;
    Print( " ]\n" );

    # the endomorphisms of the L-presentation
    endos:=EndomorphismsOfLpGroup(G);
    Print( "endomorphism = [" );
    n:=Length(endos);
    if n > 0 then 
      Print( "\n" );
      ViewObj(endos[1]);
      for i in [2..n] do 
        Print( ",\n" );
        ViewObj(endos[i]);
      od;
    fi;
    Print( " ]\n" );
  
    # iterated relators of the L-presentation
    itrels:=IteratedRelatorsOfLpGroup(G);
    Print( "iterated relators = [" );
    n:=Length(itrels);
    if n > 0 then
      Print( "\n", itrels[1] );
      for i in [2..n] do
        Print( ",\n", itrels[i]);
      od;
    fi;
    Print( " ]\n" );
end );


############################################################################
##
#M  FreeGeneratorsOfLpGroup ( G )
##
InstallMethod( FreeGeneratorsOfLpGroup,
   "for a L-presented group",
   true,
   [ IsLpGroup and IsGroupOfFamily ], 0,
   G -> GeneratorsOfGroup( FreeGroupOfLpGroup (G)));

############################################################################
##
#M  FreeGroupOfLpGroup( F ) . . underlying free group of an L-presented group
##
InstallMethod( FreeGroupOfLpGroup, 
   "for an L-presented group", 
   true,
   [ IsLpGroup and IsGroupOfFamily ], 0,
   G -> ElementsFamily( FamilyObj( G ) )!.freeGroup );


############################################################################
##
#M  FixedRelatorsOfLpGroup( F )
##
InstallMethod( FixedRelatorsOfLpGroup,
    "for L-presented groups",
    true,
    [ IsLpGroup and IsGroupOfFamily ], 0,
    G -> ElementsFamily( FamilyObj( G ) )!.relators );

############################################################################
##
#M  IteratedRelatorsOfLpGroup( F )
##
InstallMethod( IteratedRelatorsOfLpGroup,
    "for L-presented groups",
    true,
    [ IsLpGroup and IsGroupOfFamily ], 0,
    G -> ElementsFamily( FamilyObj( G ) )!.itrels );


############################################################################
##
#M  EndomorphismsOfLpGroup( F )
##
InstallMethod( EndomorphismsOfLpGroup,
    "for L-presented groups",
    true,
    [ IsLpGroup and IsGroupOfFamily ], 0,
    G -> ElementsFamily( FamilyObj( G ) )!.endos );


#############################################################################
##
#M  ViewObj(<G>)
##
InstallMethod( ViewObj,
   "L-presented group",
   true,
   [ IsLpGroup ], 0,
   function(G)
     if IsFreeGroup(G) then TryNextMethod(); fi;
     if IsGroupOfFamily(G) then
       Print("<L-presented group");
       if Length(GeneratorsOfGroup(G)) > GAPInfo.ViewLength * 10 then
         Print(" with ",Length(GeneratorsOfGroup(G))," generators>");
       else
         Print(" on the generators ",GeneratorsOfGroup(G),">");
       fi;
     else
       Error("not implemented yet\n");
     fi;
   end);

############################################################################
##
#F  LPresentedGroup ( <F> , <rels> , <Endos> , <itrels> ) 
##
InstallGlobalFunction(LPresentedGroup,
    function( F, rels, endos, itrels)
    local G,	# new object of an L-presentation
	  fam,  # new family of an L-presentation
	  gens; # generators of our group

    # Create a new family.
    fam := NewFamily( "FamilyElementsLpGroup", IsElementOfLpGroup );

    # Create the default type for the elements.
    fam!.defaultType := NewType( fam, IsPackedElementDefaultRep );

    fam!.freeGroup := F;
    fam!.relators := Immutable( rels );
    fam!.endos    := Immutable( endos );
    fam!.itrels   := Immutable( itrels );

    # Create the group.
    G := Objectify(
         NewType( CollectionsFamily( fam ),
            IsLpGroup and IsWholeFamily and IsAttributeStoringRep ),
         rec() );

    SetFilterObj(G,IsGroupOfFamily);
    
    # an ascending L-presentation is an invariant L-presentation
    if IsAscendingLPresentation(G) then
      SetIsInvariantLPresentation(G,true);
      SetEmbeddingOfAscendingSubgroup(G,GroupHomomorphismByImagesNC(G,G,
                                    GeneratorsOfGroup(G),GeneratorsOfGroup(G)));
    fi;

    # Create generators of the group.
    gens:= List( GeneratorsOfGroup( F ), g -> ElementOfLpGroup( fam, g ) );
    SetGeneratorsOfGroup( G, gens );
    if IsEmpty( gens ) then
      SetOne( G, ElementOfLpGroup( fam, One( F ) ) );
    fi;

    return G;
    end);

############################################################################
##
#P  IsAscendingLPresentationt( <G> )
##
InstallMethod( IsAscendingLPresentation,
  "for L-presented groups",
  [ IsLpGroup ], 0,
  G-> FixedRelatorsOfLpGroup(G)=[] );

############################################################################
##
#P  IsInvariantLPresentation( <G> )
##
InstallMethod( IsInvariantLPresentation,
  "for ascending L-presentations",
  [ IsLpGroup and IsAscendingLPresentation ], 0,
  G -> true);

############################################################################
##
#A  UnderlyingInvariantLPresentation( <G> )
##
InstallMethod( UnderlyingInvariantLPresentation,
  "for non-invariant L-presented groups",
  [ IsLpGroup ], 0,
  function(G)
  local H;	# the underlying invariant LpGroup

  H:=LPresentedGroup( FreeGroupOfLpGroup(G), [], EndomorphismsOfLpGroup(G),
		       IteratedRelatorsOfLpGroup(G));
  SetIsAscendingLPresentation(H,true);
  SetEmbeddingOfAscendingSubgroup(H,GroupHomomorphismByImagesNC(H,H,
                                    GeneratorsOfGroup(H),GeneratorsOfGroup(H)));
  SetIsInvariantLPresentation(H,true);
  return(H);
  end);
   
############################################################################
##
#M  EpimorphismFromFpGroup ( <LpGroup>, <n> )
##
##  returns an epimorphism from a finitely presented group G that is achieved
##  from <LpGroup> by applying the endomorphisms at most <n> times into the 
##  L-presented group <LpGroup>
##
InstallMethod( EpimorphismFromFpGroup,
  "for an L-presented group and a position integer",
  true,
  [IsLpGroup, IsPosInt], 0,
  function (L, n)
  local G, 	# the finitely presented group
  	F, 	# free group for <G>
	rels,	# relators for <G>
 	endos,	# set endomorphisms that acts on the iterated relations
	epi,	# epimorphism from <G> onto <L>
	i;	# loop variable
   
  F:=FreeGroupOfLpGroup(L);
  rels:=ShallowCopy(FixedRelatorsOfLpGroup(L));

  # determine all words in the monoid of the endomorphism of length at most <n>
  endos:=WordsOfLengthAtMostN(EndomorphismsOfLpGroup(L),n);

  # apply the endomorphisms
  for i in [1..Length(endos)] do 
    Append(rels,List(ShallowCopy(IteratedRelatorsOfLpGroup(L)),x->x^endos[i]));
  od;
  
  # the finitely presented group 
  G:=F/rels;

  # the epimorphism from <G> onto L
  epi:=GroupHomomorphismByImagesNC(G,L,GeneratorsOfGroup(G),
				       GeneratorsOfGroup(L));

  return(epi);
  end);

############################################################################
##
#M  SplitExtensionByAutomorphismsLpGroup ( <G>, <H>, <auts> )
##
##  returns the split extension of <G> by <H> where the action of <H> on 
##  <G> is given by the automorphisms <aut>: <H> -> <G> in <auts>.
##
InstallMethod(SplitExtensionByAutomorphismsLpGroup,
  "for two LpGroups and an automorphism",
  true,
  [ IsLpGroup, IsFpGroup, IsList] , 0,
  function( G, H, auts )
  local gens1,	# generators of <G> 
  	gens2,	# generators of <H> 
	F,	# free group of the split extension of <G> by <H>
	Fgens1,	# generators of the free group corr. to those of <G>
	Fgens2,	# generators of the free group corr. to those of <H>
	endo,	# an endomorphism
	map,	# MappingGeneratorsImages of <end>
	endos,	# endomorphisms of the L-presentation of the split extension
	itrels,	# iterated relators of the split extension
	rels,	# fixed relators of the split extension
	act,	# action of an automorphism
	i;	# loop variable

  gens1:=FreeGeneratorsOfLpGroup(G);
  gens2:=FreeGeneratorsOfFpGroup(H);

  # construct the free group on the union of gens1 and gens2
  F:=FreeGroup(Concatenation(List(gens1,String),List(gens2,String)));
  Fgens1:=GeneratorsOfGroup(F){[1..Length(gens1)]};
  Fgens2:=GeneratorsOfGroup(F){[Length(gens1)+1..Length(GeneratorsOfGroup(F))]};

  # build the iterated relators as union of both itrels
  itrels:=Concatenation(
          List(IteratedRelatorsOfLpGroup(G),x->MappedWord(x,gens1,Fgens1)),
          List(RelatorsOfFpGroup(H),x->MappedWord(x,gens2,Fgens2)));

  # build the new endomorphisms
  endos:=[];
  for endo in EndomorphismsOfLpGroup(G) do 
    map:=ShallowCopy(MappingGeneratorsImages(endo));
    map[1]:=List(map[1],x->MappedWord(x,gens1,Fgens1));
    map[2]:=List(map[2],x->MappedWord(x,gens1,Fgens1));
    Append(map[1],Fgens2);
    Append(map[2],Fgens2);
    Add(endos,GroupHomomorphismByImagesNC(F,F,map[1],map[2]));
  od;
  
  # build the fixed relations as union of the fixed relation and the action of
  # the automorphism
  rels:=List(FixedRelatorsOfLpGroup(G),x->MappedWord(x,gens1,Fgens1));

  # add the action of the automorphisms
  if not Length(auts)=Length(gens2) then 
    return(fail);
  fi;
  for i in [1..Length(auts)] do 
    act:=List(gens1,x->MappedWord(x,gens1,GeneratorsOfGroup(G)));
    act:=List(act,x->x^auts[i]);
    act:=List([1..Length(act)],x->Fgens1[x]^Fgens2[i]/MappedWord(act[x],
				GeneratorsOfGroup(G),Fgens1));
    Append(rels,act);
  od;

  return(LPresentedGroup(F,rels,endos,itrels));
  end);

############################################################################
##
#M  SplitExtensionByAutomorphismsLpGroup ( <G>, <H>, <auts> )
##
##  returns the split extension of <G> by <H> where the action of <H> on 
##  <G> is given by the automorphisms <aut>: <H> -> <G> in <auts>.
##
InstallMethod(SplitExtensionByAutomorphismsLpGroup,
  "for two LpGroups and an automorphism",
  true,
  [IsLpGroup,IsLpGroup,IsList], 0,
  function( G, H, auts )
  local gens1,	# generators of <G> 
  	gens2,	# generators of <H> 
	F,	# free group of the split extension of <G> by <H>
	Fgens1,	# generators of the free group corr. to those of <G>
	Fgens2,	# generators of the free group corr. to those of <H>
	endo,	# an endomorphism
	map,	# MappingGeneratorsImages of <end>
	endos,	# endomorphisms of the L-presentation of the split extension
	itrels,	# iterated relators of the split extension
	rels,	# fixed relators of the split extension
	act,	# action of an automorphism
	i;	# loop variable

  gens1:=FreeGeneratorsOfLpGroup(G);
  gens2:=FreeGeneratorsOfLpGroup(H);

  # construct the free group on the union of gens1 and gens2
  F:=FreeGroup(Concatenation(List(gens1,String),List(gens2,String)));
  Fgens1:=GeneratorsOfGroup(F){[1..Length(gens1)]};
  Fgens2:=GeneratorsOfGroup(F){[Length(gens1)+1..Length(GeneratorsOfGroup(F))]};

  # build the iterated relators as union of both itrels
  itrels:=Concatenation(
          List(IteratedRelatorsOfLpGroup(G),x->MappedWord(x,gens1,Fgens1)),
          List(IteratedRelatorsOfLpGroup(H),x->MappedWord(x,gens2,Fgens2)));

  # build the new endomorphisms
  endos:=[];
  for endo in EndomorphismsOfLpGroup(G) do 
    map:=ShallowCopy(MappingGeneratorsImages(endo));
    map[1]:=List(map[1],x->MappedWord(x,gens1,Fgens1));
    map[2]:=List(map[2],x->MappedWord(x,gens1,Fgens1));
    Append(map[1],Fgens2);
    Append(map[2],Fgens2);
    Add(endos,GroupHomomorphismByImagesNC(F,F,map[1],map[2]));
  od;
  for endo in EndomorphismsOfLpGroup(H) do 
    map:=ShallowCopy(MappingGeneratorsImages(endo));
    map[1]:=List(map[1],x->MappedWord(x,gens2,Fgens2));
    map[2]:=List(map[2],x->MappedWord(x,gens2,Fgens2));
    map[1]:=Concatenation(Fgens1,map[1]);
    map[2]:=Concatenation(Fgens1,map[2]);
    Add(endos,GroupHomomorphismByImagesNC(F,F,map[1],map[2]));
  od;
  
  # build the fixed relations as union of the fixed relation and the action of
  # the automorphism
  rels:=Concatenation(
          List(FixedRelatorsOfLpGroup(G),x->MappedWord(x,gens1,Fgens1)),
          List(FixedRelatorsOfLpGroup(H),x->MappedWord(x,gens2,Fgens2)));

  # add the action of the automorphisms
  if not Length(auts)=Length(gens2) then 
    return(fail);
  fi;
  for i in [1..Length(auts)] do 
    act:=List(gens1,x->MappedWord(x,gens1,GeneratorsOfGroup(G)));
    act:=List(act,x->x^auts[i]);
    act:=List([1..Length(act)],x->Fgens1[x]^Fgens2[i]/MappedWord(act[x],
				GeneratorsOfGroup(G),Fgens1));
    Append(rels,act);
  od;

  return(LPresentedGroup(F,rels,endos,itrels));
  end);
